from typing import Any, Callable
from flax import linen as nn
from jax import numpy as jnp
from flax.typing import Array, Initializer
from itertools import starmap
import jax

class MOE(nn.Module):
    expert_init:Callable
    expert_num:int
    expert_topk:int
    kernel_init:Initializer=nn.initializers.lecun_normal()
    @nn.compact
    def __call__(self, x:Array) -> Array:
        experts = [self.expert_init() for _ in range(self.expert_num)]

        gating = nn.Dense(self.expert_num, use_bias=False, kernel_init=self.kernel_init)(x)
        _, indices = jax.lax.top_k(gating, self.expert_topk)
        
        mask = nn.one_hot(indices, self.expert_num)
        mask = jnp.sum(mask, axis=-2)
        gating -= 1 / mask - 1
        gating = nn.softmax(gating)

        mask = jnp.split(mask, self.expert_num, axis=-1)
        gating = jnp.split(gating, self.expert_num, axis=-1)

        out = list(starmap(lambda mask, gate, expert:expert(x * mask) * gate, zip(mask, gating, experts)))
        out = sum(out)
        
        return out
    
class MultiHeadMOE(nn.Module):
    expert_init:Callable
    expert_num:int
    expert_topk:int
    head_num:int
    kernel_init:Initializer=nn.initializers.lecun_normal()
    @nn.compact
    def __call__(self, x:Array):
        assert x.shape[-1] % self.head_num == 0, "dims must be devisible by num of heads"

        dim = x.shape[-1]

        x = nn.Dense(dim, use_bias=False, kernel_init=self.kernel_init)(x)
        x = x.reshape((*x.shape[:-1], self.head_num, -1))
        x = MOE(expert_init=self.expert_init, expert_num=self.expert_num, expert_topk=self.expert_topk, kernel_init=self.kernel_init)(x)
        x = x.reshape((*x.shape[:-2], -1))
        x = nn.Dense(dim, use_bias=False, kernel_init=self.kernel_init)(x)

        return x
    
class SoftMOE(nn.Module):
    expert_init:Callable
    expert_num:int
    kernel_init:Initializer=nn.initializers.lecun_normal()
    @nn.compact
    def __call__(self, x:Array):
        experts = [self.expert_init() for _ in range(self.expert_num)]

        gating = nn.Dense(self.expert_num, use_bias=False, kernel_init=self.kernel_init)(x)
        dis = nn.softmax(gating, axis=-2)
        com = nn.softmax(gating, axis=-1)

        x = jnp.einsum("...sd,...se->...ed", x, dis)

        x = jnp.split(x, self.expert_num, axis=-2)
        x = list(starmap(lambda expert, x:expert(x), zip(experts, x)))
        x = jnp.concat(x, axis=-2)

        x = jnp.einsum("...ed,...se->...sd", x, com)

        return x



from typing import Tuple, Callable
from flax import linen as nn
from jax import numpy as jnp
from flax.typing import Array, Initializer
from typing import Optional
import jax

class MLSTMBlock(nn.Module):
    head_dim:int
    head_num:int
    ft_fn:Callable=jnp.exp
    kernel_init:Initializer=nn.initializers.lecun_normal()
    @nn.compact
    def __call__(self, x:Array, hiden_state:Optional[Tuple[Array]]=None):
        skip = x
        x = nn.LayerNorm()(x)
        skip2 = nn.Dense(x.shape[-1] * 2, kernel_init=self.kernel_init)(x)
        skip2 = nn.silu(skip2)
        x = nn.Dense(x.shape[-1] * 2, kernel_init=self.kernel_init)(x)
        q_k = nn.Conv(x.shape[-1], 4)(x)
        q_k = nn.silu(q_k)
        skip3 = q_k
        x, hiden_state = MLSTMCell(head_dim=self.head_dim, head_num=self.head_num, ft_fn=self.ft_fn, kernel_init=self.kernel_init)(x, x, x, q_k, q_k, x, hiden_state)
        x = nn.GroupNorm(self.head_num)(x)
        x += skip3
        x *= skip2
        x = nn.Dense(skip.shape[-1], kernel_init=self.kernel_init)(x)
        return x + skip, hiden_state

class MLSTMCell(nn.Module):
    head_dim:int
    head_num:int
    ft_fn:Callable=jnp.exp
    kernel_init:Initializer=nn.initializers.lecun_normal()
    @nn.compact
    def __call__(self, xi:Array, xf:Array, xo:Array, xq:Array, xk:Array, xv:Array, hiden_state:Optional[Tuple[Array]]=None):
        assert xi.shape[-1] % self.head_num == 0, "num of heads must by divisible by num of input dim"
        assert xf.shape[-1] % self.head_num == 0, "num of heads must by divisible by num of input dim"
        assert xo.shape[-1] % self.head_num == 0, "num of heads must by divisible by num of input dim"
        assert xq.shape[-1] % self.head_num == 0, "num of heads must by divisible by num of input dim"
        assert xk.shape[-1] % self.head_num == 0, "num of heads must by divisible by num of input dim"
        assert xv.shape[-1] % self.head_num == 0, "num of heads must by divisible by num of input dim"

        zero_state = (
                    jnp.zeros((*xi.shape[:-2], 1, self.head_num * self.head_dim, self.head_num * self.head_dim)),
                    jnp.ones((*xi.shape[:-2], 1, self.head_num * self.head_dim))
                    )
        
        ct, nt = zero_state if hiden_state is None else hiden_state

        xi = xi.reshape(*xi.shape[:-1], self.head_num, -1)
        xf = xf.reshape(*xf.shape[:-1], self.head_num, -1)

        wi = self.param("wi", self.kernel_init, (xi.shape[-1], self.head_num, self.head_dim))
        wf = self.param("wf", self.kernel_init, (xf.shape[-1], self.head_num, self.head_dim))

        bi = self.param("bi", self.kernel_init, (1, self.head_dim))
        bf = self.param("bf", self.kernel_init, (1, self.head_dim))

        it = jnp.einsum("...ij,jik->...ik", xi, wi) + bi
        ft = jnp.einsum("...ij,jik->...ik", xf, wf) + bf
        it = it.reshape((*it.shape[:-2], -1))
        ft = ft.reshape((*ft.shape[:-2], -1))
        ot = nn.Dense(self.head_num * self.head_dim, kernel_init=self.kernel_init)(xo)

        q = nn.Dense(self.head_num * self.head_dim, kernel_init=self.kernel_init)(xq)
        k = nn.Dense(self.head_num * self.head_dim, kernel_init=self.kernel_init)(xk)
        k = k / (self.head_num * self.head_dim)
        v = nn.Dense(self.head_num * self.head_dim, kernel_init=self.kernel_init)(xv)
        it = jnp.exp(it)
        ft = self.ft_fn(ft)
        ot = nn.sigmoid(ot)

        v *= it
        cn = v[..., None] * k[..., None, :]
        kn = it * k

        ct = jnp.concat([ct, cn], axis=-3)
        nt = jnp.concat([nt, kn], axis=-2)
        ft = jnp.concat([jnp.zeros((*ft.shape[:-2], 1, ft.shape[-1])), ft], axis=-2)

        def step(cf1, cf2):
            f1, c1 = cf1
            f2, c2 = cf2
            return f1 * f2, f2 * c1 + c2
        
        _, ct = jax.lax.associative_scan(step, [ft[..., None], ct], axis=-3)
        ct = ct[..., 1:, :, :]

        _, nt = jax.lax.associative_scan(step, [ft, nt], axis=1)
        nt = nt[..., 1:, :]

        ht = jnp.einsum("...ij,...j->...i", ct, q)
        ht /= jnp.maximum(jnp.abs(jnp.sum(nt * q, axis=-1)), 1)[..., None]
        ht *= ot
        return ht, (ct[..., -1:, :, :], nt[..., -1:, :])




from flax import linen as nn
from jax import numpy as jnp
from flax.typing import Array, Initializer
from typing import Optional
import jax

class Mamba(nn.Module):
    dim:int
    state_dim:int
    kernel_size:int=3
    is_training:bool=True
    kernel_init:Initializer=nn.initializers.lecun_normal()
    @nn.compact
    def __call__(self, x:Array, hiden_state:Optional[Array]=None):
        dim = x.shape[-1]

        x1 = nn.Dense(self.dim, kernel_init=self.kernel_init)(x)
        x2 = nn.Dense(self.dim, kernel_init=self.kernel_init)(x)

        x1 = nn.Conv(self.dim, 3, kernel_init=self.kernel_init)(x1)
        x1 = nn.silu(x1)
        x1, hiden_state = SSM(state_dim=self.state_dim, is_training=self.is_training, kernel_init=self.kernel_init)(x1, hiden_state)

        x2 = nn.silu(x2)

        x = x1 * x2
        x = nn.Dense(dim, kernel_init=self.kernel_init)(x)
        return x, hiden_state

class SSM(nn.Module):
    state_dim:int
    is_training:bool=True
    kernel_init:Initializer=nn.initializers.lecun_normal()
    @nn.compact
    def __call__(self, x:Array, hiden_state:Optional[Array]=None):
        dim = x.shape[-1]

        delta = nn.Dense(self.state_dim, use_bias=False)(x)
        a = self.param("a", self.kernel_init, (1, self.state_dim))
        b = self.param("b", self.kernel_init, (1, self.state_dim))
        delta *= a + 1j * b
        weights = jnp.exp(delta)

        x = nn.Dense(self.state_dim, use_bias=False, param_dtype=jnp.complex64, kernel_init=self.kernel_init)(x)

        def step(wx1, wx2):
            w1, x1 = wx1
            w2, x2 = wx2
            return w1 * w2, w2 * x1 + x2
        
        if self.is_training:
            _, out = jax.lax.associative_scan(step, [weights, x], axis=-2)
        else:
            if hiden_state is None:
                out = x
            else:
                out = weights * hiden_state + x
            hiden_state = out

        out = nn.Dense(dim, use_bias=False, param_dtype=jnp.complex64, kernel_init=self.kernel_init)(out)
        return jnp.real(out), hiden_state


from typing import Tuple, Callable
from flax import linen as nn
from jax import numpy as jnp
from flax.typing import Array, Initializer
from typing import Optional

class SLSTMBlock(nn.Module):
    head_dim:int
    head_num:int
    ft_fn:Callable=jnp.exp
    kernel_init:Initializer=nn.initializers.lecun_normal()
    @nn.compact
    def __call__(self, x:Array, hiden_state:Optional[Tuple[Array]]=None):
        skip = x
        x = nn.LayerNorm()(x)
        i_f = nn.Conv(x.shape[-1], 4, kernel_init=self.kernel_init)(x)
        i_f = nn.silu(i_f)
        x, hiden_state = SLSTMCell(head_dim=self.head_dim, head_num=self.head_num, ft_fn=self.ft_fn, kernel_init=self.kernel_init)(x, i_f, i_f, x, hiden_state)
        x = nn.GroupNorm(self.head_num)(x)
        x1 = nn.Dense(skip.shape[-1] * 4 // 3, kernel_init=self.kernel_init)(x)
        x2 = nn.Dense(skip.shape[-1] * 4 // 3, kernel_init=self.kernel_init)(x)
        x2 = nn.silu(x2)
        x = x1 * x2
        x = nn.Dense(skip.shape[-1], kernel_init=self.kernel_init)(x)
        return x + skip, hiden_state

class SLSTMCell(nn.Module):
    head_dim:int
    head_num:int
    ft_fn:Callable=jnp.exp
    kernel_init:Initializer=nn.initializers.lecun_normal()
    @nn.compact
    def __call__(self, xz:Array, xi:Array, xf:Array, xo:Array, hiden_state:Optional[Tuple[Array]]=None):
        assert xz.shape[-1] % self.head_num == 0, "num of heads must by divisible by num of input dim"
        assert xi.shape[-1] % self.head_num == 0, "num of heads must by divisible by num of input dim"
        assert xf.shape[-1] % self.head_num == 0, "num of heads must by divisible by num of input dim"
        assert xo.shape[-1] % self.head_num == 0, "num of heads must by divisible by num of input dim"
        
        zero_state = (
                    jnp.zeros((*xz.shape[:-2], 1, self.head_num, self.head_dim)), 
                    jnp.zeros((*xz.shape[:-2], 1, self.head_num, self.head_dim)), 
                    jnp.zeros((*xz.shape[:-2], 1, self.head_num, self.head_dim))
                    )
        
        ct, ht, nt = zero_state if hiden_state is None else hiden_state

        xz = xz.reshape(*xz.shape[:-1], self.head_num, -1)
        xi = xi.reshape(*xi.shape[:-1], self.head_num, -1)
        xf = xf.reshape(*xf.shape[:-1], self.head_num, -1)
        xo = xo.reshape(*xo.shape[:-1], self.head_num, -1)

        wz = self.param("wz", self.kernel_init, (xz.shape[-1], self.head_num, self.head_dim))
        wi = self.param("wi", self.kernel_init, (xi.shape[-1], self.head_num, self.head_dim))
        wf = self.param("wf", self.kernel_init, (xf.shape[-1], self.head_num, self.head_dim))
        wo = self.param("wo", self.kernel_init, (xo.shape[-1], self.head_num, self.head_dim))

        bz = self.param("bz", self.kernel_init, (1, self.head_dim))
        bi = self.param("bi", self.kernel_init, (1, self.head_dim))
        bf = self.param("bf", self.kernel_init, (1, self.head_dim))
        bo = self.param("bo", self.kernel_init, (1, self.head_dim))

        rz = self.param("rz", self.kernel_init, (self.head_dim, self.head_num, self.head_dim))
        ri = self.param("ri", self.kernel_init, (self.head_dim, self.head_num, self.head_dim))
        rf = self.param("rf", self.kernel_init, (self.head_dim, self.head_num, self.head_dim))
        ro = self.param("ro", self.kernel_init, (self.head_dim, self.head_num, self.head_dim))

        szt = jnp.einsum("...ij,jik->...ik", xz, wz) + bz
        sit = jnp.einsum("...ij,jik->...ik", xi, wi) + bi
        sft = jnp.einsum("...ij,jik->...ik", xf, wf) + bf
        sot = jnp.einsum("...ij,jik->...ik", xo, wo) + bo

        out = []
        for zt, it, ft, ot in zip(*list(map(lambda x:jnp.split(x, x.shape[-3], axis=-3), (szt, sit, sft, sot)))):
            zt += jnp.einsum("...ij,jik->...ik", ht, rz)
            it += jnp.einsum("...ij,jik->...ik", ht, ri)
            ft += jnp.einsum("...ij,jik->...ik", ht, rf)
            ot += jnp.einsum("...ij,jik->...ik", ht, ro)

            zt = nn.tanh(zt)
            it = jnp.exp(it)
            ft = self.ft_fn(ft)
            ot = nn.sigmoid(ot)

            ct = ft * ct + it * zt
            nt = ft * nt + it

            ht = ct / nt
            ht = ot * ht
            out.append(ht)
        
        out = jnp.concat(out, axis=-3)
        out = out.reshape(*out.shape[:-2], -1)
        return out, (ct, ht, nt)